from . import pyamber
import os
import pandas as pd
from pathlib import Path


def density():
    path = Path.cwd()
    final_file = []
    for out_file in path.glob('final_*.out'):
        final_file.append(out_file)
    input = (f"for FILE in {' '.join([item.name for item in final_file])} \n"
             "readdata \$FILE name MD \n"
             "done \n"
             "evalplateau *[Density] name EQ out Eval.agr resultsout Eval.results\n"
             "go\n"
             "quit"
             )

    with open("cpptraj.in", "w") as f:
        f.write(input)
    os.system(f'cpptraj -i cpptraj.in')
    result = pd.read_csv("Eval.results", sep="\s+")
    if result["EQ[result]"][0] == "yes":
        return 0
    elif result["EQ[result]"][0] == "no":
        return 1
    else:
        raise RuntimeError("Equil failed!")


def prep(rst7, s, temp, heavymask, backbonemask, loop=20):
    min1 = pyamber.Min("step1", systemInfo=s, ref=rst7,
                       refc=rst7, restraintmask=heavymask, restraint_wt=5.0)
    min1.Run()
    nvt1 = pyamber.NVT("step2", systemInfo=s, ref="step1.rst7", refc="step1.rst7", temp=temp,
                       restraintmask=heavymask, restraint_wt=5.0, nstlim=15000, tautp=0.5)
    nvt1.Run()
    min2 = pyamber.Min('step3', systemInfo=s, ref='step2.rst7',
                       refc='step2.rst7', restraintmask=heavymask, restraint_wt=2.0)
    min2.Run()
    min3 = pyamber.Min('step4', systemInfo=s, ref='step3.rst7',
                       refc='step3.rst7', restraintmask=backbonemask, restraint_wt=0.1)
    min3.Run()
    min4 = pyamber.Min('step5', systemInfo=s,
                       ref='step4.rst7', refc='step4.rst7')
    min4.Run()
    npt1 = pyamber.NPT("step6", systemInfo=s, ref="step5.rst7", temp=temp,
                       refc="step5.rst7", restraintmask=heavymask, restraint_wt=1.0)
    npt1.Run()
    npt2 = pyamber.NPT("step7", systemInfo=s, ref="step6.rst7", refc="step5.rst7", temp=temp,
                       irest=True, restraintmask=heavymask, restraint_wt=0.5)
    npt2.Run()
    npt3 = pyamber.NPT("step8", systemInfo=s, ref="step7.rst7", refc="step5.rst7", temp=temp,
                       irest=True, restraintmask=backbonemask, restraint_wt=0.5, nstlim=10000)
    npt3.Run()
    npt4 = pyamber.NPT("step9", systemInfo=s, ref="step8.rst7", temp=temp,
                       refc="step5.rst7", irest=True, dt=0.002, nscm=1000)
    npt4.Run()

    ref = "step9.rst7"
    for i in range(loop):
        final = pyamber.NPT(f"final_{i}", systemInfo=s, ref=ref, temp=temp,
                            refc="step5.rst7", irest=True, dt=0.002, nscm=1000, nstlim=500000, ntwx=5000)
        final.Run()
        result = density()
        if result == 0:
            return f'final_{i}.rst7'
        ref = f'final_{i}.rst7'
    raise RuntimeError(
        "More than 20 iterations of final density equil required. Bailing out.")




import pytest

import env
from pybind11_tests import ConstructorStats
from pybind11_tests import modules as m
from pybind11_tests.modules import subsubmodule as ms


def test_nested_modules():
    import pybind11_tests

    assert pybind11_tests.__name__ == "pybind11_tests"
    assert pybind11_tests.modules.__name__ == "pybind11_tests.modules"
    assert (
        pybind11_tests.modules.subsubmodule.__name__
        == "pybind11_tests.modules.subsubmodule"
    )
    assert m.__name__ == "pybind11_tests.modules"
    assert ms.__name__ == "pybind11_tests.modules.subsubmodule"

    assert ms.submodule_func() == "submodule_func()"


def test_reference_internal():
    b = ms.B()
    assert str(b.get_a1()) == "A[1]"
    assert str(b.a1) == "A[1]"
    assert str(b.get_a2()) == "A[2]"
    assert str(b.a2) == "A[2]"

    b.a1 = ms.A(42)
    b.a2 = ms.A(43)
    assert str(b.get_a1()) == "A[42]"
    assert str(b.a1) == "A[42]"
    assert str(b.get_a2()) == "A[43]"
    assert str(b.a2) == "A[43]"

    astats, bstats = ConstructorStats.get(ms.A), ConstructorStats.get(ms.B)
    assert astats.alive() == 2
    assert bstats.alive() == 1
    del b
    assert astats.alive() == 0
    assert bstats.alive() == 0
    assert astats.values() == ["1", "2", "42", "43"]
    assert bstats.values() == []
    assert astats.default_constructions == 0
    assert bstats.default_constructions == 1
    assert astats.copy_constructions == 0
    assert bstats.copy_constructions == 0
    # assert astats.move_constructions >= 0  # Don't invoke any
    # assert bstats.move_constructions >= 0  # Don't invoke any
    assert astats.copy_assignments == 2
    assert bstats.copy_assignments == 0
    assert astats.move_assignments == 0
    assert bstats.move_assignments == 0


def test_importing():
    from collections import OrderedDict

    from pybind11_tests.modules import OD

    assert OD is OrderedDict
    assert str(OD([(1, "a"), (2, "b")])) == "OrderedDict([(1, 'a'), (2, 'b')])"


def test_pydoc():
    """Pydoc needs to be able to provide help() for everything inside a pybind11 module"""
    import pydoc

    import pybind11_tests

    assert pybind11_tests.__name__ == "pybind11_tests"
    assert pybind11_tests.__doc__ == "pybind11 test module"
    assert pydoc.text.docmodule(pybind11_tests)


def test_duplicate_registration():
    """Registering two things with the same name"""

    assert m.duplicate_registration() == []


def test_builtin_key_type():
    """Test that all the keys in the builtin modules have type str.

    Previous versions of pybind11 would add a unicode key in python 2.
    """
    if hasattr(__builtins__, "keys"):
        keys = __builtins__.keys()
    else:  # this is to make pypy happy since builtins is different there.
        keys = __builtins__.__dict__.keys()

    assert {type(k) for k in keys} == {str}


@pytest.mark.xfail("env.PYPY", reason="PyModule_GetName()")
def test_def_submodule_failures():
    sm = m.def_submodule(m, b"ScratchSubModuleName")  # Using bytes to show it works.
    assert sm.__name__ == m.__name__ + "." + "ScratchSubModuleName"
    malformed_utf8 = b"\x80"
    if env.PYPY:
        # It is not worth the effort finding a trigger for a failure when running with PyPy.
        pytest.skip("Sufficiently exercised on platforms other than PyPy.")
    else:
        # Meant to trigger PyModule_GetName() failure:
        sm_name_orig = sm.__name__
        sm.__name__ = malformed_utf8
        try:
            # We want to assert that a bad __name__ causes some kind of failure, although we do not want to exercise
            # the internals of PyModule_GetName(). Currently all supported Python versions raise SystemError. If that
            # changes in future Python versions, simply add the new expected exception types here.
            with pytest.raises(SystemError):
                m.def_submodule(sm, b"SubSubModuleName")
        finally:
            # Clean up to ensure nothing gets upset by a module with an invalid __name__.
            sm.__name__ = sm_name_orig  # Purely precautionary.
    # Meant to trigger PyImport_AddModule() failure:
    with pytest.raises(UnicodeDecodeError):
        m.def_submodule(sm, malformed_utf8)



from . import _pyamber


class SystemInfo(_pyamber.SystemInfo):
    def __init__(self, parm7, rst7, runMin="pmemd.cuda_DPFP", runMd="pmemd.cuda") -> None:
        super().__init__(prmtop=parm7, inpcrd=rst7, runMin=runMin, runMd=runMd)

    getNprotein = _pyamber.SystemInfo.getNprotein
    getnDna = _pyamber.SystemInfo.getnDna
    getnRna = _pyamber.SystemInfo.getnRna
    getnLipid = _pyamber.SystemInfo.getnLipid
    getnUnKnown_ = _pyamber.SystemInfo.getnUnKnown_
    getnCharmmWater_ = _pyamber.SystemInfo.getnCharmmWater_
    getnWater = _pyamber.SystemInfo.getnWater
    getnCarbo = _pyamber.SystemInfo.getnCarbo
    getHasCharmmWater = _pyamber.SystemInfo.getHasCharmmWater
    getHasOrthoBox = _pyamber.SystemInfo.getHasOrthoBox
    getRunMin = _pyamber.SystemInfo.getRunMin
    getRunMd = _pyamber.SystemInfo.getRunMd
    getHeavyMask = _pyamber.SystemInfo.getHeavyMask
    getBackBoneMask = _pyamber.SystemInfo.getBackBoneMask


class Min(_pyamber.Min):
    def __init__(self, name, systemInfo, ref, refc,
                 irest=False, restraintmask="", restraint_wt=0.0, cut=8.0, nTmin=2, maxCyc=1000,
                 nCyc=10, nTwx=500, nTpr=50, nTwr=500) -> None:
        super().__init__(name, systemInfo, ref, refc, irest=irest, restraintmask=restraintmask,
                         restraint_wt=restraint_wt, cut=cut, nTmin=nTmin, maxCyc=maxCyc, nCyc=nCyc, nTwx=nTwx,
                         nTpr=nTpr, nTwr=nTwr)

    setCut = _pyamber.Min.setCut
    setNTpr = _pyamber.Min.setNTpr
    setNTwr = _pyamber.Min.setNTwr
    setNTwx = _pyamber.Min.setNTwx
    setMaxCyc = _pyamber.Min.setMaxCyc
    setNCyc = _pyamber.Min.setNCyc
    setNTim = _pyamber.Min.setNTim
    Run = _pyamber.Min.Run


class NVT(_pyamber.Nvt):
    def __init__(self, name, systemInfo, ref, refc,
                 irest=False, temp=303.15, restraintmask="", restraint_wt=0.0, nstlim=5000, cut=8.0, ntb=1,
                 ntc=2, ntf=2, tautp=1.0, gamma_ln=5.0, dt=0.002, nscm=0, ntwx=500, ntpr=50, ntwr=500) -> None:
        super().__init__(name, systemInfo, ref, refc, irest=irest, temp=temp, restraintmask=restraintmask,
                         restraint_wt=restraint_wt,
                         nstlim=nstlim, cut=cut, ntc=ntc, ntf=ntf, tautp=tautp, gamma_ln=gamma_ln, dt=dt, nscm=nscm,
                         ntwx=ntwx, ntpr=ntpr, ntwr=ntwr)

    Run = _pyamber.Nvt.Run
    setCut = _pyamber.Nvt.setCut
    setTemp = _pyamber.Nvt.setTemp
    setNTpr = _pyamber.Nvt.setNTpr
    setNTwx = _pyamber.Nvt.setNTwx
    setNTwr = _pyamber.Nvt.setNTwr
    setNstLim = _pyamber.Nvt.setNstLim
    setIrest = _pyamber.Nvt.setIrest
    setTautp = _pyamber.Nvt.setTautp
    setGammaLn = _pyamber.Nvt.setGammaLn
    setDt = _pyamber.Nvt.setDt
    setNscm = _pyamber.Nvt.setNscm
    setNtx = _pyamber.Nvt.setNtx
    setNtc = _pyamber.Nvt.setNtc
    setNtf = _pyamber.Nvt.setNtf
    setThermostat = _pyamber.Nvt.setThermostat


class NPT(_pyamber.Npt):
    def __init__(self, name, systemInfo, ref, refc,
                 irest=False, temp=303.15, restraintmask="", restraint_wt=0.0, nstlim=5000, cut=8.0, ntc=2, ntf=2,
                 tautp=1.0, taup=1.0, mcbarint=100, gamma_ln=5.0, dt=0.002, nscm=0, ntwx=500, ntpr=50,
                 ntwr=500) -> None:
        super().__init__(name, systemInfo, ref, refc, irest=irest, temp=temp, restraintmask=restraintmask,
                         restraint_wt=restraint_wt,
                         nstlim=nstlim, cut=cut, ntc=ntc, ntf=ntf, tautp=tautp, taup=taup, mcbarint=mcbarint,
                         gamma_ln=gamma_ln, dt=dt, nscm=nscm, ntwx=ntwx, ntpr=ntpr, ntwr=ntwr)

    Run = _pyamber.Npt.Run
    setCut = _pyamber.Npt.setCut
    setTemp = _pyamber.Npt.setTemp
    setNTpr = _pyamber.Npt.setNTpr
    setNTwx = _pyamber.Npt.setNTwx
    setNTwr = _pyamber.Npt.setNTwr
    setNstLim = _pyamber.Npt.setNstLim
    setIrest = _pyamber.Npt.setIrest
    setTautp = _pyamber.Npt.setTautp
    setTaup = _pyamber.Npt.setTaup
    setMcbarint = _pyamber.Npt.setMcbarint
    setGammaLn = _pyamber.Npt.setGammaLn
    setDt = _pyamber.Npt.setDt
    setNscm = _pyamber.Npt.setNscm
    setNtx = _pyamber.Npt.setNtx
    setNtc = _pyamber.Npt.setNtc
    setNtf = _pyamber.Npt.setNtf
    setThermostat = _pyamber.Npt.setThermostat


class GaMd(_pyamber.GaMd):
    def __init__(self, name, systemInfo, ref, refc,
                 irest=False, temp=303.15, restraintmask="", restraint_wt=0.0, nstlim=5000, cut=8.0, ntc=2, ntf=2,
                 tautp=1.0, taup=1.0, mcbarint=100, gamma_ln=5.0, dt=0.002, nscm=0, ntwx=500, ntpr=50, ntwr=500,
                 igamd=1, ie=1, iep=1, ied=1, ntcmdprep=200000, ntcmd=1000000, ntebprep=200000, nteb=1000000,
                 ntave=50000, irest_gamd=0, sigma0P=6.0, sigma0D=6.0, timask1="", scmask1="") -> None:
        super().__init__(name, systemInfo, ref, refc, irest=irest, temp=temp, restraintmask=restraintmask,
                         restraint_wt=restraint_wt,
                         nstlim=nstlim, cut=cut, ntc=ntc, ntf=ntf, tautp=tautp, taup=taup, mcbarint=mcbarint,
                         gamma_ln=gamma_ln, dt=dt, nscm=nscm, ntwx=ntwx, ntpr=ntpr, ntwr=ntwr,
                         igamd=igamd, ie=ie, iep=iep, ied=ied, ntcmdprep=ntcmdprep, ntcmd=ntcmd, ntebprep=ntebprep,
                         nteb=nteb, ntave=ntave, irest_gamd=irest_gamd, sigma0P=sigma0P, sigma0D=sigma0D,
                         timask1=timask1, scmask1=scmask1)

    Run = _pyamber.GaMd.Run
    setCut = _pyamber.GaMd.setCut
    setTemp = _pyamber.GaMd.setTemp
    setNTpr = _pyamber.GaMd.setNTpr
    setNTwx = _pyamber.GaMd.setNTwx
    setNTwr = _pyamber.GaMd.setNTwr
    setNstLim = _pyamber.GaMd.setNstLim
    setIrest = _pyamber.GaMd.setIrest
    setTautp = _pyamber.GaMd.setTautp
    setTaup = _pyamber.GaMd.setTaup
    setMcbarint = _pyamber.GaMd.setMcbarint
    setGammaLn = _pyamber.GaMd.setGammaLn
    setDt = _pyamber.GaMd.setDt
    setNscm = _pyamber.GaMd.setNscm
    setNtx = _pyamber.GaMd.setNtx
    setNtc = _pyamber.GaMd.setNtc
    setNtf = _pyamber.GaMd.setNtf
    setThermostat = _pyamber.GaMd.setThermostat
    setIGaMd = _pyamber.GaMd.setIGaMd
    setIe = _pyamber.GaMd.setIe
    setIep = _pyamber.GaMd.setIep
    setIed = _pyamber.GaMd.setIed
    setNtcmdprep = _pyamber.GaMd.setNtcmdprep
    setNteb = _pyamber.GaMd.setNteb
    setNtave = _pyamber.GaMd.setNtave
    setIrest_gamd = _pyamber.GaMd.setIrest_gamd
    setSigma0P = _pyamber.GaMd.setSigma0P
    setSigmaOD = _pyamber.GaMd.setSigmaOD
    setTimask1 = _pyamber.GaMd.setTimask1
    setTimask1 = _pyamber.GaMd.setTimask1



import argparse
from . import pyamber
from . import equil


def arg_parse():
    parser = argparse.ArgumentParser(description='Tools for automated operation of AMBER MD')
    parser.add_argument('--parm7', '-p', type=str,
                        required=True, help="amber top file")
    parser.add_argument('--rst7', '-c', type=str,
                        required=True, help='amber rst file')
    parser.add_argument('--temp', '-t', type=float,
                        required=False, help='Temperature', default=298.15)
    parser.add_argument("--ns", '-n', type=int,
                        help="time for MD(ns)", default=100)
    parser.add_argument('--addmask', default=None,
                        type=str, help="add restarint mask")
    parser.add_argument("--gamd", type=bool, default=False, help="if run gamd")
    parser.add_argument("--MIN", type=str, default="pmemd.cuda_DPFP", help="Engine for MIN")
    parser.add_argument("--MD", type=str, default="pmemd.cuda", help="Engine for MD")
    args = parser.parse_args()
    return args


def main():
    args = arg_parse()
    parm7 = args.parm7
    rst7 = args.rst7
    temp = args.temp
    gamd = args.gamd
    s = pyamber.SystemInfo(parm7, rst7, runMin=args.MIN, runMd=args.MD)
    ns = args.ns
    if args.addmask is not None:
        if len(s.getBackBoneMask()) > 0 and len(s.getHeavyMask()) > 0:
            heavymask = "\"" + s.getHeavyMask() + "|" + args.addmask + "\""
            backbonemask = "\"" + s.getBackBoneMask() + "|" + args.addmask + "\""
        else:
            heavymask = "\"" + args.addmask + "\""
            backbonemask = "\"" + args.addmask + "\""
    else:
        heavymask = "\"" + s.getHeavyMask() + "\""
        backbonemask = "\"" + s.getBackBoneMask() + "\""
    rst7 = equil.prep(rst7=rst7, s=s, temp=temp, heavymask=heavymask,
                      backbonemask=backbonemask, loop=20)
    if gamd:
        md = pyamber.GaMd("GaMd", s, rst7, rst7, irest=True,
                          nscm=1000, nstlim=ns * 500000, ntwx=50000)
        md.Run()
    else:
        md = pyamber.NPT("md", s, rst7, rst7, ntwx=50000, irest=True, nscm=1000, nstlim=ns * 500000)
        md.Run()


if __name__ == '__main__':
    main()
